/////////////////////////////////////////////////////////////////////
// UVA TeamName RoboCup agent developed for the Simulator
// League of the RoboCup 2001 competition.
//
// This team based on the Mainz Rolling Brains RoboCup team from
// the 2000 competition.
//
// Copyright (C) 2001 Team Members
// of the UVA TeamName RoboCup Team.
//
// This program under the terms of the GNU General Public License as
// published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
// GoalieInterceptperception.cpp Tue Jul 3 2001
// Description:
//////////////////////////////////////////////////////////////////////

#include "goalieinterceptperception.h"

#undef TRACE
#define TRACE(cmd)

GoalieInterceptPerception::GoalieInterceptPerception(const WorldModel &_wmWorld) : Perception(_wmWorld)
{}

GoalieInterceptPerception::~GoalieInterceptPerception()
{}

// deprecating this function for needToIntercept function

bool GoalieInterceptPerception::needToIntercept() const
{
  // needToIntercept returns true if the player closest to the ball
  // is not at the ball and if the client player is either the closest
  // or second closest player to the ball.
  
  if (goalieHasBall())
    return false;
  
  WorldPlayer wpPlayer;
  ClientWorldPlayer cwpClientPlayer= wmWorld.getClientPlayer();
  WorldBall wbBall = wmWorld.getWorldBall();

  // if the client player cannot see the ball he cannot intercept it
  if (wbBall.getAge() > 10)
    return false;
  
  if (ballInPenaltyArea() && (dist(wbBall, cwpClientPlayer) < 15.0))
    return true;
  
  if (dist(wbBall, cwpClientPlayer) < 7.0)
    return true;
  
  if ((wbBall.getVelocity ().getX () < -1.0)
      && wbBall.getAbsPosition ().getX () < -30.0)
    {
      return true;
    }

  return false;
}

bool GoalieInterceptPerception::atBall(const WorldPlayer &_wpPlayer) const
{
  WorldBall wbBall = wmWorld.getWorldBall();
  return (dist(_wpPlayer, wbBall) < (0.7 * KICKRANGE));
}

Angle GoalieInterceptPerception::angleToBall() const
{
  WorldBall wbBall = wmWorld.getWorldBall();
  Vector2d vRelPos = relPos(wbBall);
  
  return vRelPos.getAngle();
}

Vector2d GoalieInterceptPerception::accelToIntercept(int _iTime) const
{
  // this function computes the acceleration necessary to intercept the ball at a given time.
  // the formula's used come from solving differential equations that model player and ball
  // motion as described in the server manual. (player_position(t) = ball_position(t) where
  // position''(t) = position'(t) * decay + accel_constant

  if (_iTime == 0)
    {
      return Vector2d(0.0, 0.0);
    }

  Vector2d vAcceleration;
  ClientWorldPlayer cwpClientPlayer = wmWorld.getClientPlayer();
  WorldBall wbBall = wmWorld.getWorldBall();
  
  Vector2d vPlayerInitialPos = cwpClientPlayer.getAbsPosition();
  Vector2d vPlayerInitialVel = cwpClientPlayer.getVelocity();
  Vector2d vBallInitialPos = wbBall.getAbsPosition();
  Vector2d vBallInitialVel = wbBall.getVelocity();
  

  double Kp = (1 - pow(PLAYER_DECAY, _iTime + 1)) / (1 - PLAYER_DECAY);
  double Kb = (1 - pow(BALL_DECAY, _iTime + 1)) / (1 - BALL_DECAY);
    //	double dDenominator = (_iTime + 1)/(1 - PLAYER_DECAY) - PLAYER_DECAY * Kp;
  double dDenominator = 1;
  
  vAcceleration = ((vBallInitialPos - vPlayerInitialPos)/Kp + vBallInitialVel * Kb/Kp) / dDenominator - vPlayerInitialVel;
  
  
  //	INSPECT("B0: " << vBallInitialPos);
  //	INSPECT("P0: " << vPlayerInitialPos);
  //	INSPECT("b0: " << vBallInitialVel);
  //	INSPECT("v0: " << vPlayerInitialVel);
  //	INSPECT("ball_decay: " << BALL_DECAY);
  //	INSPECT("player_decay: " << PLAYER_DECAY);
  //	INSPECT("Kp: " << Kp);
  //	INSPECT("Kb: " << Kb);
  //	INSPECT("acc: " << vAcceleration.getLength() << ", " << vAcceleration.getAngle());
  
  
  //	double dDenominator = (1/(PLAYER_DECAY*PLAYER_DECAY))*(exp(-PLAYER_DECAY*_iTime) - 1 + PLAYER_DECAY * _iTime);
  
  //	vAcceleration = (vBallInitialPos + (1/BALL_DECAY) * ( 1 - exp(-BALL_DECAY * (_iTime))) * vBallInitialVel - vPlayerInitialPos
  //											- (1/PLAYER_DECAY) * (1 - exp(-PLAYER_DECAY * _iTime)) * vPlayerInitialVel) / dDenominator;
  
  
  // finally rotate the vector to be aligned with the player's coordinate system
  vAcceleration.setRadAng(vAcceleration.getLength(), vAcceleration.getAngle() - cwpClientPlayer.getBodyDir());
  
  return vAcceleration;
}

bool GoalieInterceptPerception::closestOpponentToBall(WorldPlayer &_wpPlayer) const
{
    WorldPlayer wpPlayer;
    WorldBall wbBall = wmWorld.getWorldBall();
    double dDist;
    double dShortestDist = INFINITY;
    int iClosestJerseyNum = UNKNOWN;
    int i;

    for(i = 1; i <= MAX_PLAYER; i++)
    {
        wpPlayer = wmWorld.getOtherPlayer(i);
        if (wpPlayer.isKnown())
        {
            dDist = dist(wpPlayer, wbBall);
            if (dDist < dShortestDist)
            {
                dShortestDist = dDist;
                iClosestJerseyNum = wpPlayer.getNumber();
            }
        }
    }

    if (iClosestJerseyNum != UNKNOWN)
    {
        _wpPlayer = wmWorld.getOtherPlayer(iClosestJerseyNum);
        return true;
    }
    else
    {
        return false;
    }
}

bool GoalieInterceptPerception::isTooFarFromBall() const
{
    WorldBall wbBall = wmWorld.getWorldBall();
    Vector2d vBallPos = wbBall.getAbsPosition();

    return (relPos(vBallPos).getLength() >= 35.0);
}

bool GoalieInterceptPerception::intercept(Vector2d &_vInterceptPoint, int &_iTime, const Vector2d& _vInitialPos,
					  const Vector2d& _vInitialVel, double _dAcceleration, int playerDelay = 1) const
{
  // the computations in this function are taken directly out of the server manual.
  // originally this function was meant to apply a closed form differential equation,
  // however, upon investigating the optimization problem no closed formed solution was found.
  // Hence, the function simulates movement in a method similar to the server
  
  double dPlayerPos = 0;														// the radius of a circle of future positions centered at the players current position
  double dBallPos = INFINITY;												// the magnitude of the distance between the ball and the client players position
  WorldBall wbBall = wmWorld.getWorldBall();				// the ball
  _iTime = -wbBall.getAge();																				// relative time (in server cycles) from the current time
  Vector2d vBallVelocity = wbBall.getVelocity();		//the ball's current velocity
  Vector2d vBallPos = wbBall.getAbsPosition();    	// absolute position of ball on field
  double dPlayerSpeed = _vInitialVel.getLength();		// the speed of the player on the field
  bool fWillRunPast = false;
  
  // all players have a max acceleration, do not pass that acceleration
  // TODO: get rid of magic number 1.0 with PLAYER_ACCEL_MAX constant
  if (_dAcceleration > 1.0)
    {
      TRACE("Attempted to accelerate at too high of a value, " << _dAcceleration << " changing to PLAYER_ACCEL_MAX.");
      _dAcceleration = 1.0;
    }
  
  // loop until the player's possible position and the ball's possible position overlap or until the player
  // passes the ball (because he's running too fast)
  while (!( ((dPlayerPos < (dBallPos + wbBall.getSize()/2))  &&
	     (dPlayerPos > (dBallPos - wbBall.getSize()/2)))
	    || fWillRunPast ))
    {
      // since it takes the player one turn to move, assume that we lose the first timestep
      if (_iTime >= playerDelay)
    	{
            // add the player's acceleration to the players velocity ( vf = v0 + a*t where t = 1)
            dPlayerSpeed = dPlayerSpeed + _dAcceleration;
            // ensure that the player does not exceed max velocity
            if (dPlayerSpeed > PLAYER_SPEED_MAX)
            {
                dPlayerSpeed = PLAYER_SPEED_MAX;
            }

            // add the players position to the players speed (s = s0 + v*t where t = 1)
            dPlayerPos += dPlayerSpeed;

            // apply a decay factor to the speed because of friction
            dPlayerSpeed = dPlayerSpeed * PLAYER_DECAY;
        }

        // compute the ball's new position using the velocity
        vBallPos = vBallPos + vBallVelocity;
        // compute the relative distance of the ball from the original position of the player
        dBallPos = norm(vBallPos - _vInitialPos);

        // adjust the ball's velocity in proportion to the decay factor
        vBallVelocity = vBallVelocity * BALL_DECAY;
        // if the ball has stopped and we're passed it, we ran through the ball
        fWillRunPast = (fabs(vBallVelocity.getLength()) < 0.1f) && (dPlayerPos > dBallPos);

        // increase to the next time step
        _iTime++;
    }

    if(fWillRunPast) INSPECT("running past the ball.. damnit!");
    INSPECT(getTime() + _iTime << "~: expect ball to be at: " << vBallPos
         << " replay command: c" << vBallPos.getX() << "," << vBallPos.getY() << ",1");
    INSPECT(getTime() << ": client replay command: c" << _vInitialPos.getX() << ","
         << _vInitialPos.getY() << "," << dPlayerPos);


    // if the client player is going to run past the ball or if the ball's final position is unrealistic..
    if (fWillRunPast || (dBallPos > FIELD_LENGTH))
    {
        return false;
    }

    // otherwise, we intercept the ball at the ball's last predicted position
    _vInterceptPoint = vBallPos;

    return true;

}

Angle GoalieInterceptPerception::accelAngleToIntercept(double _dDashPower) const
{

    Vector2d vAbsTargetPos;						// where the client player needs to be to intercept the ball with the given dash power
    int iTargetTime;								// relative time (in server cycles) from the current time when the player intercepts the ball
    double dAccelMagnitude;					// the magnitude of the acceleration of the client player
    bool fWillIntercept;						// bool representing whether or not the client player will intercept
    ClientWorldPlayer cwpClientPlayer = wmWorld.getClientPlayer();	// the client player
	WorldBall wbBall = wmWorld.getWorldBall();

    // formula for converting dash->acceleration from server manual
    dAccelMagnitude = cwpClientPlayer.getEffort() * DASH_POWER_RATE * _dDashPower;
    fWillIntercept = intercept(vAbsTargetPos, iTargetTime, cwpClientPlayer.getAbsPosition(), cwpClientPlayer.getVelocity(), dAccelMagnitude);


    /*
    	Vector2d vAcceleration = accelToIntercept(iTargetTime);
    	double dMagnitude = vAcceleration.getLength();
    	Angle aFormula = vAcceleration.getAngle();
    	Angle aRelFormula = vAcceleration.getAngle() - cwpClientPlayer.getFaceDir();
    	Angle aRelTarget = relPos(vAbsTargetPos).getAngle();
    	Angle aAbsTarget = vAbsTargetPos.getAngle();
    */

    if (fWillIntercept) // && wbBall.getVelocity().getLength() >= 0.2)
    {
        // the relative target position is the difference between the balls predicted position
        // and the player's current position
//        cerr << getTime() << ":bodydir: " << RAD_TO_DEG(cwpClientPlayer.getBodyDir()) << " velocity: " <<
//            cwpClientPlayer.getVelocity().getLength() << endl;

        double x0, x1, y0, y1;
        x0 = vAbsTargetPos.getX();
        y0 = vAbsTargetPos.getY();
        x1 = cwpClientPlayer.getAbsPosition().getX();
        y1 = cwpClientPlayer.getAbsPosition().getY();

        INSPECT (getTime() << ": Trajectory to intersect: l" << x0 << "," << y0
		 << "," << x1 << "," << y1);


//        cerr << getTime() << "vAbsTargetPos: " << vAbsTargetPos << endl;
        Angle aResult = relPos(vAbsTargetPos).getAngle();

        Vector2d v;
        v.setRadAng(10.0, aResult + cwpClientPlayer.getBodyDir());


//        cerr << getTime() << " actual angle: " << RAD_TO_DEG(aResult) << endl;


        return aResult;
        //return aRelFormula;
    }
    else
    {
//        if(fWillIntercept) cerr << "fWillIntercept";
//        else cerr << "!!!fWillIntercept";
//        cerr << " Ball is slow, or !fWillIntercept" << endl;
//        TRACE("The client player is either trying to move so fast that he will pass the ball or thinks the ball is very far away.");
        // set the target position equal to the balls current position
        return angleToBall();
    }
}

bool GoalieInterceptPerception::isActiveMode() const
{
    play_mode pmTemp = wmWorld.getPlayMode();
    if( (pmTemp == PLAY_MODE_PLAY_ON) ||
            (pmTemp == PLAY_MODE_DROP_BALL) ||
            (pmTemp == PLAY_MODE_KICK_OFF_US) ||
            (pmTemp == PLAY_MODE_FREE_KICK_US) ||
            (pmTemp == PLAY_MODE_KICK_IN_US) ||
            (pmTemp == PLAY_MODE_CORNER_KICK_US) ||
            (pmTemp == PLAY_MODE_OFFSIDE_THEM) ||  // Not sure about this one
            (pmTemp == PLAY_MODE_GOAL_KICK_US)
      )
        return true;

    return false;
}

bool GoalieInterceptPerception::specialSituation() const
{
    play_mode pmTemp = wmWorld.getPlayMode();
    if(     (pmTemp == PLAY_MODE_KICK_OFF_US) ||
            (pmTemp == PLAY_MODE_FREE_KICK_US) ||
            (pmTemp == PLAY_MODE_KICK_IN_US) ||
            (pmTemp == PLAY_MODE_CORNER_KICK_US) ||
            (pmTemp == PLAY_MODE_GOAL_KICK_US)
      ) return true;

    return false;
}


int GoalieInterceptPerception::getDashVector(Vector2d &_vDash) const
{
  WorldPlayer wpOpponent;
  ClientWorldPlayer cwpClientPlayer = wmWorld.getClientPlayer();
  WorldBall wbBall = wmWorld.getWorldBall();
  bool fWillOpponentIntercept;
  bool fWillClientIntercept;
  Vector2d vAbsInterceptPos;
  Vector2d vAcceleration;
  Vector2d vAbsTargetPos;
  int iTime;
  double dDashPower;
  double predictDash = 60.0;
  double dAccelMagnitude = cwpClientPlayer.getEffort() 
    * DASH_POWER_RATE * predictDash;
  fWillClientIntercept = intercept 
    (vAbsTargetPos, iTime, 
     cwpClientPlayer.getAbsPosition(),
     cwpClientPlayer.getVelocity(), 
     dAccelMagnitude);

  INSPECT ("My position:  " << cwpClientPlayer.getAbsPosition ()
	   << " velocity: " << cwpClientPlayer.getVelocity ());
  INSPECT ("Current ball: " << wbBall.getAbsPosition ()
	   << " velocity: " << cwpClientPlayer.getVelocity ());
  INSPECT ("Intercepting in " << iTime << " at " << vAbsTargetPos);

  Angle aFastIntercept = accelAngleToIntercept (predictDash);
  _vDash.setRadAng(100, aFastIntercept);
  
  return iTime;
}		

# if 0  
  // if there are no opponents going for the ball, go towards the ball at 50
  if  (!closestOpponentToBall(wpOpponent))
    {
      Angle aSlowIntercept = accelAngleToIntercept(50);
      _vDash.setRadAng(50, aSlowIntercept);
      
      return INFINITY;  // !@#need to return a big int
    }


    Vector2d vAbsTargetPos;
    int iTargetTime;
    double dAccelMagnitude = cwpClientPlayer.getEffort() * DASH_POWER_RATE * 100.0;

    fWillClientIntercept = intercept(vAbsTargetPos, iTargetTime, cwpClientPlayer.getAbsPosition(), cwpClientPlayer.getVelocity(), dAccelMagnitude);
    // results in vAbsInterceptPos & iTime
    fWillOpponentIntercept = intercept(vAbsInterceptPos, iTime, wpOpponent.getAbsPosition(), wpOpponent.getVelocity(), 1.0);
    int iOpponentDifTime = iTime - 5;

    if (iTime < iTargetTime)
    {
      // if the opponent is visible but won't get to the ball any time soon, go towards the ball at 50
      // uhh?
      if (!fWillOpponentIntercept)
      {
          cerr << "No opponenet will intercept ball" << endl;
          Angle aSlowIntercept = accelAngleToIntercept(50);
          _vDash.setRadAng(50, aSlowIntercept);

          return INFINITY; //@!#
      }

      vAcceleration = accelToIntercept(iTime);
      dDashPower = vAcceleration.getLength() / (cwpClientPlayer.getEffort() * DASH_POWER_RATE);

    }
    else  // if our time is less, go get it on our circle
    {

        vAcceleration = accelToIntercept(max(iTargetTime, iOpponentDifTime));
        dDashPower = vAcceleration.getLength() / (cwpClientPlayer.getEffort() * DASH_POWER_RATE);
    }

    if (dDashPower > 100)
    {
        cerr << "dDashPower too great: " << dDashPower << endl;
        Angle aFastIntercept = accelAngleToIntercept(100);
        _vDash.setRadAng(100, aFastIntercept);

          return max(iTargetTime, iOpponentDifTime);
    }
    else if (dDashPower < 50)
    {
        cerr << "dDashPower less than 50: " << dDashPower << endl;
        Angle aSlowIntercept = accelAngleToIntercept(50);
        _vDash.setRadAng(50, aSlowIntercept);
    }
    else
    {
        cerr << "dDashPower is: " << dDashPower << endl;
        _vDash.setRadAng(dDashPower, vAcceleration.getAngle());
    }

    return max(iTargetTime, iOpponentDifTime);
}

# endif

bool GoalieInterceptPerception::ballInPenaltyArea() const
{
    Vector2d vAbsPos = wmWorld.getWorldBall().getAbsPosition();

    return ( ( vAbsPos.getX() > - 0.5 * FIELD_LENGTH ) &&
             ( vAbsPos.getX() < - 0.5 * FIELD_LENGTH + PENALTY_AREA_LENGTH ) &&
             ( fabs( vAbsPos.getY() ) < 0.5 * PENALTY_AREA_WIDTH ) );

}

int GoalieInterceptPerception::getWorldAge() const
{
	WorldBall wbBall = wmWorld.getWorldBall();
	
	return wbBall.getAge();		
}

bool GoalieInterceptPerception::sendDashOnly() const
{
	return wmWorld.getTime() != wmWorld.getVisualTimeStamp();
}

bool GoalieInterceptPerception::inDefensiveHalf(const WorldObject& _woObj) const
{
	//gets the absolute position of the WMO
	Vector2d vAbsPos = _woObj.getAbsPosition();
	
	//finds if the X coord is less than -1/6 Length because less than that
	//is the defensive third of the field
	if ( vAbsPos.getX() < 0 )
		return true;
	else return false;
}

bool GoalieInterceptPerception::goalieHasBall() const
{
    WorldPlayer wpGoalie = wmWorld.getOurGoalie();
    WorldBall wbBall = wmWorld.getWorldBall();

    if (!wpGoalie.isKnown())
        return false;

    if (wpGoalie.getAge() > 5.0)
        return false;

    if (dist(wpGoalie, wbBall) < 1.0)
        return true;

    return false;
}
